#include "scripting/lua-bindings/manual/CCLuaEngine.h"
#include "AdjoeDelegate.h"

namespace adjoe {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)

    struct AdjoeParams {
        const char *UANetwork;
        const char *UAChannel;
        const char *UASubPublisherCleartext;
        const char *UASubPublisherEncrypted;
        const char *Placement;
    };

    struct Options {
        const char *userId;
        AdjoeParams params;
    };

    static inline AdjoeParams getAdjoeParams(lua_State *L) {
        AdjoeParams params = {};

        // Assume Table on Top of Stack
        lua_getfield(L, -1, "UANetwork");
        if (lua_isstring(L, -1)) {
            params.UANetwork = lua_tostring(L, -1);
        }
        lua_pop(L, 1);

        lua_getfield(L, -1, "UAChannel");
        if (lua_isstring(L, -1)) {
            params.UAChannel = lua_tostring(L, -1);
        }
        lua_pop(L, 1);

        lua_getfield(L, -1, "UASubPublisherCleartext");
        if (lua_isstring(L, -1)) {
            params.UASubPublisherCleartext = lua_tostring(L, -1);
        }
        lua_pop(L, 1);

        lua_getfield(L, -1, "UASubPublisherEncrypted");
        if (lua_isstring(L, -1)) {
            params.UASubPublisherEncrypted = lua_tostring(L, -1);
        }
        lua_pop(L, 1);

        lua_getfield(L, -1, "Placement");
        if (lua_isstring(L, -1)) {
            params.Placement = lua_tostring(L, -1);
        }
        lua_pop(L, 1);
        return params;
    }

    static inline Options getOptions(lua_State *L) {
        Options options = {};

        // Assume Table on Top of Stack
        lua_getfield(L, -1, "AdjoeParams");
        if (lua_istable(L, -1)) {
            options.params = getAdjoeParams(L);
        }
        lua_pop(L, 1);

        lua_getfield(L, -1, "UserID");
        if (lua_isstring(L, -1)) {
            options.userId = lua_tostring(L, -1);
        }
        lua_pop(L, 1);
        return options;
    }

    static inline jobject constructJavaOptions(JNIEnv *env, AdjoeParams params) {
        jclass optionsClass = env->FindClass("io/adjoe/cocos2dx/AdjoeOptions");
        jmethodID optionConstructor = env->GetMethodID(optionsClass, "<init>", "()V");
        jobject adjoeOptions = env->NewObject(optionsClass, optionConstructor);

        if (params.UANetwork != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setUaNetwork",
                                                      "(Ljava/lang/String;)V");
            jstring uaNetwork = env->NewStringUTF(params.UANetwork);
            env->CallVoidMethod(adjoeOptions, setUaNetwork,
                                uaNetwork);
        }
        if (params.UAChannel != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setUaChannel",
                                                      "(Ljava/lang/String;)V");
            jstring uaChannel = env->NewStringUTF(params.UAChannel);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, uaChannel);
        }
        if (params.UASubPublisherCleartext != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setUaSubPublisherCleartext",
                                                      "(Ljava/lang/String;)V");
            jstring uaSubPublisherCleartext = env->NewStringUTF(params.UASubPublisherCleartext);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, uaSubPublisherCleartext);
        }
        if (params.UASubPublisherEncrypted != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass,
                                                      "setUaSubPublisherEncrypted",
                                                      "(Ljava/lang/String;)V");
            jstring uaSubPublisherEncrypted = env->NewStringUTF(params.UASubPublisherEncrypted);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, uaSubPublisherEncrypted);
        }
        if (params.Placement != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setPlacement",
                                                      "(Ljava/lang/String;)V");
            jstring placement = env->NewStringUTF(params.Placement);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, placement);
        }
        return adjoeOptions;
    }

    static inline jobject constructJavaOptions(JNIEnv *env, Options options) {
        jclass optionsClass = env->FindClass("io/adjoe/cocos2dx/AdjoeOptions");
        jmethodID optionConstructor = env->GetMethodID(optionsClass, "<init>", "()V");
        jobject adjoeOptions = env->NewObject(optionsClass, optionConstructor);

        if (options.userId != nullptr) {
            jmethodID mSetUserId = env->GetMethodID(optionsClass, "setUserId",
                                                    "(Ljava/lang/String;)V");
            jstring userId = env->NewStringUTF(options.userId);
            env->CallVoidMethod(adjoeOptions, mSetUserId, userId);
        }
        if (options.params.UANetwork != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setUaNetwork",
                                                      "(Ljava/lang/String;)V");
            jstring uaNetwork = env->NewStringUTF(options.params.UANetwork);
            env->CallVoidMethod(adjoeOptions, setUaNetwork,
                                uaNetwork);
        }
        if (options.params.UAChannel != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setUaChannel",
                                                      "(Ljava/lang/String;)V");
            jstring uaChannel = env->NewStringUTF(options.params.UAChannel);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, uaChannel);
        }
        if (options.params.UASubPublisherCleartext != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass,
                                                      "setUaSubPublisherCleartext",
                                                      "(Ljava/lang/String;)V");
            jstring uaSubPublisherCleartext = env->NewStringUTF(
                    options.params.UASubPublisherCleartext);
            env->CallVoidMethod(adjoeOptions, setUaNetwork,
                                uaSubPublisherCleartext);
        }
        if (options.params.UASubPublisherEncrypted != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass,
                                                      "setUaSubPublisherEncrypted",
                                                      "(Ljava/lang/String;)V");
            jstring uaSubPublisherEncrypted = env->NewStringUTF(
                    options.params.UASubPublisherEncrypted);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, uaSubPublisherEncrypted);
        }
        if (options.params.Placement != nullptr) {
            jmethodID setUaNetwork = env->GetMethodID(optionsClass, "setPlacement",
                                                      "(Ljava/lang/String;)V");
            jstring placement = env->NewStringUTF(options.params.Placement);
            env->CallVoidMethod(adjoeOptions, setUaNetwork, placement);
        }
        return adjoeOptions;
    }

    static JNIEnv *getJNIEnv() {
        static JavaVM *jvm = cocos2d::JniHelper::getJavaVM();
        JNIEnv *jniEnv;

        int jvmEnvErr = jvm->GetEnv((void **) &jniEnv, JNI_VERSION_1_6);

        if (jvmEnvErr == JNI_EDETACHED) {
            if (jvm->AttachCurrentThread(&jniEnv, nullptr) != 0) {
                return nullptr;
            }
            jvmEnvErr = jvm->GetEnv((void **) &jniEnv, JNI_VERSION_1_6);
        }
        if (jvmEnvErr == JNI_OK) {
            return jniEnv;
        }
        return nullptr;
    }

    static void DEBUG(const char *msg) {
        __android_log_print(ANDROID_LOG_ERROR, "AdjoeDelegateJNI", "%s\n", msg);
    }

#endif

    int adjoe_init(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }
        jstring sdkHash = env->NewStringUTF(lua_tostring(L, 1));
        int failureCallback = luaL_ref(L, LUA_REGISTRYINDEX);
        int successCallback = luaL_ref(L, LUA_REGISTRYINDEX);

        jobject initOptions = nullptr;

        if (lua_istable(L, -1)) {
            Options options = getOptions(L);
            initOptions = constructJavaOptions(env, options);
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID initMethod = env->GetStaticMethodID(adjoeModule, "initAdjoe",
                                                      "(Ljava/lang/String;Lio/adjoe/cocos2dx/AdjoeOptions;II)V");

        env->CallStaticVoidMethod(adjoeModule, initMethod, sdkHash, initOptions, successCallback,
                                  failureCallback);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoec_is_initialized(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID isInitializedMethod = env->GetStaticMethodID(adjoeModule, "isInitialized",
                                                               "()Z");
        jboolean canShow = env->CallStaticBooleanMethod(adjoeModule, isInitializedMethod);
        lua_pushboolean(L, (int) canShow);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_show_ow(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID showOfferwallMethod = env->GetStaticMethodID(adjoeModule, "showOfferwall",
                                                               "(Lio/adjoe/cocos2dx/AdjoeOptions;)Ljava/lang/String;");

        jobject params = constructJavaOptions(env, getAdjoeParams(L));
        auto errorString = (jstring) env->CallStaticObjectMethod(adjoeModule, showOfferwallMethod,
                                                                 params);
        if (errorString == nullptr) {
            return 0;
        }

        const char *nativeErrorString = env->GetStringUTFChars(errorString, nullptr);
        lua_pushstring(L, nativeErrorString);
        env->ReleaseStringUTFChars(errorString, nativeErrorString);
        lua_error(L);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_set_profile(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID setProfileMethod = env->GetStaticMethodID(adjoeModule, "setProfile",
                                                            "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Lio/adjoe/cocos2dx/AdjoeOptions;)Ljava/lang/String;");

        jobject params = constructJavaOptions(env, getAdjoeParams(L));

        const char *source_str = lua_tostring(L, 1);
        const char *gender_str = lua_tostring(L, 2);
        const char *birthday_str = lua_tostring(L, 3);

        jstring source = source_str != nullptr ? env->NewStringUTF(source_str) : nullptr;
        jstring gender = gender_str != nullptr ? env->NewStringUTF(gender_str) : nullptr;
        jstring birthday = birthday_str != nullptr ? env->NewStringUTF(birthday_str) : nullptr;

        auto errorString = (jstring) env->CallStaticObjectMethod(adjoeModule, setProfileMethod,
                                                                 source, gender, birthday, params);
        if (errorString == nullptr) {
            return 0;
        }

        const char *nativeErrorString = env->GetStringUTFChars(errorString, nullptr);
        lua_pushstring(L, nativeErrorString);
        env->ReleaseStringUTFChars(errorString, nativeErrorString);
        lua_error(L);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_can_show_ow(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID canShowOfferwallMethod = env->GetStaticMethodID(adjoeModule, "canShowOfferwall",
                                                                  "()Z");
        jboolean canShow = env->CallStaticBooleanMethod(adjoeModule, canShowOfferwallMethod);
        lua_pushboolean(L, (int) canShow);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_get_version(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID getVersionMethod = env->GetStaticMethodID(adjoeModule, "getVersion",
                                                            "()I");
        jint version = env->CallStaticIntMethod(adjoeModule, getVersionMethod);
        lua_pushinteger(L, (int) version);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_get_version_name(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID canShowOfferwallMethod = env->GetStaticMethodID(adjoeModule, "getVersionName",
                                                                  "()Ljava/lang/String;");
        auto versionName = (jstring) env->CallStaticObjectMethod(adjoeModule,
                                                                 canShowOfferwallMethod);
        const char *nativeVersionName = env->GetStringUTFChars(versionName, nullptr);
        lua_pushstring(L, nativeVersionName);
        env->ReleaseStringUTFChars(versionName, nativeVersionName);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_has_accepted_tos(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID hasAcceptedTosMethod = env->GetStaticMethodID(adjoeModule, "hasAcceptedTOS",
                                                                "()Z");
        jboolean hasAcceptedTos = env->CallStaticBooleanMethod(adjoeModule, hasAcceptedTosMethod);
        lua_pushboolean(L, (int) hasAcceptedTos);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_has_accepted_usage_permission(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID hasAcceptedUPMethod = env->GetStaticMethodID(adjoeModule,
                                                               "hasAcceptedUsagePermission",
                                                               "()Z");
        jboolean hasAcceptedUsagePerm = env->CallStaticBooleanMethod(adjoeModule,
                                                                     hasAcceptedUPMethod);
        lua_pushboolean(L, (int) hasAcceptedUsagePerm);
        return 1;
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_request_rewards(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        int failureCallback = luaL_ref(L, LUA_REGISTRYINDEX);
        int successCallback = luaL_ref(L, LUA_REGISTRYINDEX);
        jobject params = constructJavaOptions(env, getAdjoeParams(L));
        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID initMethod = env->GetStaticMethodID(adjoeModule, "requestRewards",
                                                      "(Lio/adjoe/cocos2dx/AdjoeOptions;II)V");
        env->CallStaticVoidMethod(adjoeModule, initMethod, params, successCallback,
                                  failureCallback);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_send_user_event(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID sendEventMethod = env->GetStaticMethodID(adjoeModule, "sendEvent",
                                                           "(ILio/adjoe/cocos2dx/AdjoeOptions;)Ljava/lang/String;");
        jobject params = constructJavaOptions(env, getAdjoeParams(L));
        int event = lua_tointeger(L, 1);
        auto errorString = (jstring) env->CallStaticObjectMethod(adjoeModule, sendEventMethod,
                                                                 event,
                                                                 params);
        if (errorString == nullptr) {
            return 0;
        }
        const char *nativeErrorString = env->GetStringUTFChars(errorString, nullptr);
        lua_pushstring(L, nativeErrorString);
        env->ReleaseStringUTFChars(errorString, nativeErrorString);
        lua_error(L);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    int adjoe_do_payout(lua_State *L) {
#if (CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID)
        JNIEnv *env = getJNIEnv();

        if (env == nullptr) {
            return 0;
        }

        jclass adjoeModule = env->FindClass("io/adjoe/cocos2dx/AdjoeSDKModule");
        jmethodID doPayoutMethod = env->GetStaticMethodID(adjoeModule, "doPayout",
                                                          "(Lio/adjoe/cocos2dx/AdjoeOptions;II)V");

        int failureCallback = luaL_ref(L, LUA_REGISTRYINDEX);
        int successCallback = luaL_ref(L, LUA_REGISTRYINDEX);
        jobject params = constructJavaOptions(env, getAdjoeParams(L));

        env->CallStaticVoidMethod(adjoeModule, doPayoutMethod, params,
                                  successCallback, failureCallback);
#endif
        // this is a no-op on non-android platforms.
        return 0;
    }

    void bind_adjoe_lua_functions(lua_State *L) {
        lua_register(L, "adjoec_init", adjoe_init);
        lua_register(L, "adjoec_is_initialized", adjoec_is_initialized);
        lua_register(L, "adjoec_show_ow", adjoe_show_ow);
        lua_register(L, "adjoec_can_show_ow", adjoe_can_show_ow);
        lua_register(L, "adjoec_get_version", adjoe_get_version);
        lua_register(L, "adjoec_get_version_name", adjoe_get_version_name);
        lua_register(L, "adjoec_has_accepted_tos", adjoe_has_accepted_tos);
        lua_register(L, "adjoec_has_accepted_usage_permission",
                     adjoe_has_accepted_usage_permission);
        lua_register(L, "adjoec_request_rewards", adjoe_request_rewards);
        lua_register(L, "adjoec_set_profile", adjoe_set_profile);
        lua_register(L, "adjoec_send_user_event", adjoe_send_user_event);
        lua_register(L, "adjoec_do_payout", adjoe_do_payout);
    }
}
